open Syntax

val prioritysalt : int32 option ref
val enablecongclos : bool ref
val apply_funcrules_poseqns : bool ref
val self_matings : bool ref
val self_confrontations : bool ref
val ordered_shallow_rules : bool ref
val ordered_shallow_rules_strict : bool ref
val ordered_shallow_rules_salt : int32 ref

val clauses : int list list ref
val clause_assumption : (int list,unit) Hashtbl.t
val clause_princ : (int list,int list) Hashtbl.t
val clause_inst : (int list,tm) Hashtbl.t

val rote_strict : bool ref
val eigen : (string,string) Hashtbl.t
val eigensasinsts : bool ref
val pos_rote_shallow : (string,unit) Hashtbl.t
val neg_rote_shallow : (string,unit) Hashtbl.t
val pos_rote_pos : (string,unit) Hashtbl.t
val pos_rote_neg : (string,unit) Hashtbl.t
val pos_rote_inst : (string,unit) Hashtbl.t
val neg_rote_pos : (string,unit) Hashtbl.t
val neg_rote_neg : (string,unit) Hashtbl.t
val neg_rote_inst : (string,unit) Hashtbl.t
val pos_watch_shallow : (string,unit) Hashtbl.t
val neg_watch_shallow : (string,unit) Hashtbl.t
val pos_watch_pos : (string,unit) Hashtbl.t
val pos_watch_neg : (string,unit) Hashtbl.t
val pos_watch_inst : (string,unit) Hashtbl.t
val neg_watch_pos : (string,unit) Hashtbl.t
val neg_watch_neg : (string,unit) Hashtbl.t
val neg_watch_inst : (string,unit) Hashtbl.t
val shallowatom : (int,unit) Hashtbl.t
val funcdef : (string,tm) Hashtbl.t
val preddef : (string,tm) Hashtbl.t
val tmid_schema : (string,string * string list) Hashtbl.t
val pos_src_shallow : (string,string) Hashtbl.t
val pos_src_pos : (string,string) Hashtbl.t
val pos_src_neg : (string,string) Hashtbl.t
val pos_src_inst : (string,string) Hashtbl.t
val neg_src_shallow : (string,string) Hashtbl.t
val neg_src_pos : (string,string) Hashtbl.t
val neg_src_neg : (string,string) Hashtbl.t
val neg_src_inst : (string,string) Hashtbl.t
val inst_src_neg : (string,string) Hashtbl.t
val shallow_src_pos : (string,string) Hashtbl.t
val shallow_src_neg : (string,string) Hashtbl.t

type shallowrule =
  | ShallowFunc of int list * int * tm * tm * tm list
  | ShallowPred of int list * int * tm * bool * tm list

val pinv_subst : int -> tm list -> tm list * tm list

val use_quant_cases : bool ref
val use_quant_cases_pre : bool ref
val quant_cases : (tm * bool * (tm * bool) list * (int * (bool * tm) list * tm) list) list ref
val use_shallowrules_pred : bool ref
val use_shallowrules_pred_func : bool ref
val use_shallowrules_func : bool ref
val use_shallowrules_func_func : bool ref
val shallow_var_elim : bool ref
val dynamic_shallowrules : bool ref

val bvars_spine : tm list -> int list
val complete_bvars_p : int list -> int -> bool
    
val prop_as_shallow_rules : int -> string -> tm -> bool -> (string * shallowrule) list

val register_shallow_rule : string * shallowrule -> unit
val reset_shallow_rules : unit -> unit

val tm_delta_norm : tm -> tm
    
val log_const_p : string -> bool
val tm_schema_and_instnames : tm -> tm * string list
val tm_schema : tm -> tm

exception Timeout

val set_timer : float -> unit

exception AbstractTimeout

val abstract_time : int ref
val abstract_time_limit : int ref
val middle_abstract_time_limit : int ref
val prop_endphase : bool ref

exception PfFound

type searchoption =
    ProcessPosProp of string * tm
  | ProcessNegProp of string * tm
  | ProcessInst of string * tm

val atomh : (string,int) Hashtbl.t

val pos_prop_priority_1 : string -> tm -> int
val neg_prop_priority_1 : string -> tm -> int
val possibly_new_pos_prop : tm -> string list -> string list -> string list -> string list -> unit
val possibly_new_neg_prop : tm -> string list -> string list -> string list -> string list -> unit
val possibly_new_inst : tm -> string -> unit
val possibly_new_inst_init : tm -> unit
val possibly_new_pos_prop_priority : tm -> int -> unit
val possibly_new_neg_prop_priority : tm -> int -> unit
val possibly_new_inst_priority : tm -> int -> unit

val invert_negprop : tm -> tm
val negprop : tm -> tm

val shallowclause : (int list,string) Hashtbl.t
val atom_count : int ref
val atomh_rev : (int,string) Hashtbl.t
val atomh_sch_rev : (int,string) Hashtbl.t
val new_atomh : string -> int
val new_atom : tm -> int

val reset_atoms : unit -> unit
val reset_clauses : unit -> unit
val reset_search : unit -> unit

val search_main : unit -> unit

val srule_prop : shallowrule -> tm
