(defvar *dir* (directory-namestring (car (directory "."))))

(defvar *predcnt* 0)
(defvar *funccnt* 0)
(defvar *axcnt* 0)
(defvar *axnum* (make-hash-table :test #'equal))
(defvar *prednum* (make-hash-table :test #'equal))
(defvar *funcnum* (make-hash-table :test #'equal))
(defvar *predar* (make-hash-table :test #'equal))
(defvar *funcar* (make-hash-table :test #'equal))

(defvar *incomplete* nil)

(defun funtp (n r)
  (if (> n 0)
      (format nil "set -> ~d" (funtp (- n 1) r))
    r))

(defun process-fof (w &optional bvl)
  (if (consp w)
      (case (car w)
	    (ALL
	     (process-fof (caddr w) (append (cadr w) bvl)))
	    (EX
	     (process-fof (caddr w) (append (cadr w) bvl)))
	    (NOT (process-fof (cadr w) bvl))
	    ((AND OR EQUIV IMP) (process-fof (cadr w) bvl) (process-fof (caddr w) bvl))
	    (EQ (process-trm (cadr w) bvl) (process-trm (caddr w) bvl))
	    (t
	     (let ((p (car w)))
	       (cond ((equal p "$true") nil)
		     ((equal p "$false") nil)
		     (t
		      (unless (stringp p)
			(throw 'ERROR nil))
		      (if (gethash p *prednum*)
			  (unless (equal (length (cdr w)) (gethash p *predar*))
			    (format t "arity problem with ~d~%" p)
			    (throw 'ERROR NIL))
			(progn
			  (setf (gethash p *prednum*) (incf *predcnt*))
			  (setf (gethash p *predar*) (length (cdr w)))))
		      (process-spine (cdr w) bvl))))))
    (progn
      (setq *incomplete* t)
      (throw 'incomplete nil))))

(defun process-trm (m bvl)
  (if (consp m)
      (let ((f (car m)))
	(unless (stringp f)
	  (format t "~d does not look like a function or constant to me~%" f)
	  (throw 'ERROR nil))
	(if (gethash f *funcnum*)
	    (unless (equal (length (cdr m)) (gethash f *funcar*))
	      (format t "arity problem with ~d~%" f)
	      (throw 'ERROR NIL))
	  (progn
	    (setf (gethash f *funcnum*) (incf *funccnt*))
	    (setf (gethash f *funcar*) (length (cdr m)))))
	(process-spine (cdr m) bvl))
    (if m
	nil
      (throw 'incomplete nil))))

(defun process-spine (sp bvl)
  (if sp
      (progn
	(process-trm (car sp) bvl)
	(process-spine (cdr sp) bvl))
    t))

(defun bvl-discr (bvl)
  (if bvl
      (format nil " ~d~d" (car bvl) (bvl-discr (cdr bvl)))
    ""))

(defun fof-discr (w &optional bvl)
  (if (consp w)
      (case (car w)
	    (ALL
	     (format nil "(forall~d, ~d)"
		     (bvl-discr (cadr w))
		     (fof-discr (caddr w) (append (cadr w) bvl))))
	    (EX
	     (format nil "(exists~d, ~d)"
		     (bvl-discr (cadr w))
		     (fof-discr (caddr w) (append (cadr w) bvl))))
	    (NOT (format nil "(~~ ~d)" (fof-discr (cadr w) bvl)))
	    (AND (format nil "(~d /\\ ~d)" (fof-discr (cadr w) bvl) (fof-discr (caddr w) bvl)))
	    (OR (format nil "(~d \\/ ~d)" (fof-discr (cadr w) bvl) (fof-discr (caddr w) bvl)))
	    (IMP (format nil "(~d -> ~d)" (fof-discr (cadr w) bvl) (fof-discr (caddr w) bvl)))
	    (EQUIV (format nil "(~d <-> ~d)" (fof-discr (cadr w) bvl) (fof-discr (caddr w) bvl)))
	    (EQ (format nil "~d = ~d" (trm-discr (cadr w) bvl nil) (trm-discr (caddr w) bvl nil)))
	    (t
	     (let ((p (car w)))
	       (cond ((equal p "$true") "True")
		     ((equal p "$false") "False")
		     (t
		      (unless (gethash p *prednum*) (throw 'ERROR nil))
		      (format nil "p~d~d" (gethash p *prednum*) (spine-discr (cdr w) bvl)))))))
    (progn
      (format t "Translation problem~%")
      (throw 'ERROR nil))))

(defun trm-discr (m bvl &optional (par t))
  (if (consp m)
      (let ((f (car m)))
	(unless (gethash f *funcnum*) (throw 'ERROR nil))
	(if (cdr m)
	    (format nil "~df~d~d~d" (if par "(" "") (gethash f *funcnum*) (spine-discr (cdr m) bvl) (if par ")" ""))
	  (format nil "f~d" (gethash f *funcnum*))))
    (format nil "~d" m)))

(defun spine-discr (sp bvl)
  (if sp
      (format nil " ~d~d"
	      (trm-discr (car sp) bvl)
	      (spine-discr (cdr sp) bvl))
    ""))

(defun problem-axioms (f)
  (let ((l nil)
	(nm nil)
	(role nil)
	(sl nil)
	(al nil))
    (catch 'incomplete
      (loop while (setq l (read-line f nil nil)) do
	    (if (and (> (length l) 1) (eq (aref l 0) #\())
		(if (and (> (length l) 3) (equal (subseq l (- (length l) 3)) "!!!"))
		    (progn
		      (setq *incomplete* t) (throw 'incomplete nil))
		  (progn
		    (setq l (read-from-string l))
		    (cond ((and (consp l) (eq (car l) 'BIGSTOP))
			   (setq *incomplete* t))
			  ((and (consp l) (eq (car l) 'NAME))
			   (setq nm (cadr l)))
			  ((and (consp l) (member (car l) '(AXIOM QUESTION CONJECTURE NEGCONJECTURE HYPOTHESIS LEMMA WATCHLIST UNKNOWN)))
			   (setq role (car l)))
			  ((and nm role)
			   (process-fof l)
			   (push (list nm role l) al))
			  (t
			   (format t "WARNING: Unexpected ~d~%" l)))))
	      (format t "~d~%" l))))
    (close f)
    al))
  
(defun tmlimit ()
  (let ((n (member "-t" sb-ext:*posix-argv* :test #'equal)))
    (if n
	(read-from-string (cadr n))
      300)))

(defun probfile ()
  (car (last sb-ext:*posix-argv*)))

(defvar *verbosity* 1)

(defun setverbosity ()
  (let ((v (member "-v" sb-ext:*posix-argv* :test #'equal)))
    (when v
      (setq *verbosity* (read-from-string (cadr v))))))

(defun transltofile ()
  (let ((v (member "-trto" sb-ext:*posix-argv* :test #'equal)))
    (when v (cadr v))))

(defvar *starttm* 0)

(defun top ()
  (setq *starttm* (get-universal-time))
  (setverbosity)
  (let* ((tm (tmlimit))
	 (probfile (probfile)))
    (with-timeout tm
      (let ((ep (sb-ext:run-program (format nil "~d/E/PROVER/eprover" *dir*) (list "-Q" "--cpu-limit=1" probfile) :output :stream)))
	(when (equal (process-exit-code ep) 172) (setq *incomplete* t))
	(catch 'ERROR
	(let ((al (unless *incomplete*
		    (problem-axioms (process-output ep)))))
	    (let ((gn (transltofile)))
	      (when (and *incomplete* (> *verbosity* 5)) (format t "INCOMPLETE~%"))
	      (if gn
		  (unless *incomplete*
		    (let ((g (open gn :direction :output :if-exists :supersede :if-does-not-exist :create)))
		      (maphash
		       #'(lambda (p v)
			   (format g "Variable p~d: ~d.~%" v (funtp (gethash p *predar*) "prop")))
		       *prednum*)
		      (maphash
		       #'(lambda (f v)
			   (format g "Variable f~d: ~d.~%" v (funtp (gethash f *funcar*) "set")))
		       *funcnum*)
		      (let ((conj nil))
			(dolist (a al)
			  (if (eq (cadr a) 'CONJECTURE)
			      (setq conj a)
			    (when (member (cadr a) '(AXIOM NEGCONJECTURE HYPOTHESIS LEMMA))
			      (setf (gethash (car a) *axnum*) (incf *axcnt*))
			      (format g "Hypothesis a~d: ~d.~%" *axcnt* (fof-discr (caddr a))))))
			(if conj
			    (format g "Theorem conj: ~d.~%Admitted.~%" (fof-discr (caddr conj)))
			  (format g "Theorem conj: False.~%Admitted.~%"))
			(close g))))
		(let ((remtm (- tm (- (get-universal-time) *starttm*))))
		  (if (<= remtm 0)
		      (format t "Timeout~%")
		    (if (or *incomplete* (> (length al) 63)) ; call E
			(sb-ext:run-program (format nil "~d/E/PROVER/eprover" *dir*) (list "--auto-schedule" (format nil "--cpu-limit=~d" remtm) probfile) :output :stream)
		      (let* ((dp (sb-ext:run-program (format nil "~d/discriminator/bin/discriminator" *dir*) (list "-I" (format nil "~d/discriminator/FOPreamble.mgs" *dir*) "-t" (format nil "~d" remtm) "-") :input :stream :wait nil :output t))
			     (dpi (process-input dp)))
			(maphash
			 #'(lambda (p v)
			     (when (> *verbosity* 5)
			       (format t "Variable p~d: ~d.~%" v (funtp (gethash p *predar*) "prop")))
			     (format dpi "Variable p~d: ~d.~%" v (funtp (gethash p *predar*) "prop")))
			 *prednum*)
			(maphash
			 #'(lambda (f v)
			     (when (> *verbosity* 5)
			       (format t "Variable f~d: ~d.~%" v (funtp (gethash f *funcar*) "set")))
			     (format dpi "Variable f~d: ~d.~%" v (funtp (gethash f *funcar*) "set")))
			 *funcnum*)
			(let ((conj nil))
			  (dolist (a al)
			    (if (eq (cadr a) 'CONJECTURE)
				(setq conj a)
			      (when (member (cadr a) '(AXIOM NEGCONJECTURE HYPOTHESIS LEMMA))
				(setf (gethash (car a) *axnum*) (incf *axcnt*))
				(when (> *verbosity* 5)
				  (format t "Hypothesis a~d: ~d.~%" *axcnt* (fof-discr (caddr a))))
				(format dpi "Hypothesis a~d: ~d.~%" *axcnt* (fof-discr (caddr a))))))
			  (when (> *verbosity* 5)
			    (if conj
				(format t "Theorem conj: ~d.~%Admitted.~%" (fof-discr (caddr conj)))
			      (format t "Theorem conj: False.~%Admitted.~%")))
			  (if conj
			      (format dpi "Theorem conj: ~d.~%Admitted.~%" (fof-discr (caddr conj)))
			    (format dpi "Theorem conj: False.~%Admitted.~%"))
			  (close dpi)
			  (process-wait dp))))))))))))))

(setf sb-ext:*invoke-debugger-hook* #'(lambda(&rest args) (sb-ext:exit)))

(sb-ext:save-lisp-and-die "epidemic" :toplevel #'top :executable t)

  
